import json
import datetime

class CustomJSONEncoder(json.JSONEncoder):

	def default(obj):
		if type(obj) is datetime.date:
			return str(obj)  # use iso format
		elif hasattr(obj, '__dataclass_fields__'):
			return vars(obj)  # used for serializing dataclasses
		else:
			raise TypeError



# Tests for vim_input module
# We use a simple representation of the
# VimEditor which only shows text and pos
# cursor:
## [ represents normal mode
## | represents insert mode
'''
>>> v
[The quick brown fox jumped over the lazy dog.
>>> # MOTION
>>> ## Simple Motions
>>> v.process_keys('$'); v
The quick brown fox jumped over the lazy dog.[
>>> v.process_keys('0'); v
[The quick brown fox jumped over the lazy dog.
>>> v.process_keys('w'); v
The [quick brown fox jumped over the lazy dog.
>>> v.process_keys('2w'); v
The quick brown [fox jumped over the lazy dog.
>>> v.process_keys('b'); v
The quick [brown fox jumped over the lazy dog.
>>> v.process_keys('5b'); v
[The quick brown fox jumped over the lazy dog.
>>> v.process_keys('3e'); v
The quick brow[n fox jumped over the lazy dog.
>>> v.process_keys('2h'); v
The quick br[own fox jumped over the lazy dog.
>>> v.process_keys('4l'); v
The quick brown [fox jumped over the lazy dog.
>>> ## Search Motions
>>> v.process_keys('fj'); v
The quick brown fox [jumped over the lazy dog.
>>> v.process_keys('tz'); v
The quick brown fox jumped over the l[azy dog.
>>> v.process_keys('Tv'); v
The quick brown fox jumped ov[er the lazy dog.
>>> v.process_keys('Ff'); v
The quick brown [fox jumped over the lazy dog.
>>> # ACTIONS
>>> v.process_keys('rb'); v # replace
The quick brown [box jumped over the lazy dog.
>>> v.process_keys('llsok '+ESC); v # sub
The quick brown book[  jumped over the lazy dog.
>>> v.process_keys(ESC + 'x'); v # delete char
The quick brown book[ jumped over the lazy dog.
>>> v.process_keys('5X'); v # backspace
The quick brown[ jumped over the lazy dog.
>>> v.process_keys('u'); v # backspace
The quick brown book[ jumped over the lazy dog.
>>> v.process_keys('2u'); v # backspace
The quick brown [box jumped over the lazy dog.
>>> v.process_keys('Sall new'+ESC); v # substitution
all ne[w
>>> v.process_keys('u'); v # backspace
The quick brown [box jumped over the lazy dog.
>>> v.process_keys('4wD'); v # delete (to end of line)
The quick brown box jumped over the[ 
>>> v.process_keys('2bCinto the crate'+ESC); v # change ( to end )
The quick brown box jumped into the crat[e
>>> v.process_keys('2bY'); v # yank (to end of line)
The quick brown box jumped into [the crate
>>> v.yank
'the crate'
>>> v.process_keys('$p'); v # paste
The quick brown box jumped into the crate[the crate
>>> # INSERTION ACTIONS
>>> # i I a A
>>> v.process_keys('i under'+ESC); v
The quick brown box jumped into the crate unde[rthe crate
>>> v.process_keys('a '+ESC); v
The quick brown box jumped into the crate under[ the crate
>>> v.process_keys('Iprefix '+ESC); v
prefix[ The quick brown box jumped into the crate under the crate
>>> v.process_keys('A suffix'+ESC); v
prefix The quick brown box jumped into the crate under the crate suffi[x
>>> v.line_history = ['Something came before','in line_history']
>>> v.process_keys('2k'); v
Something came befor[e
>>> v.process_keys('j'); v
in line_histor[y
>>> v.process_keys('j'); v
[
>>> v.process_keys('2k'); v
Something came befor[e
>>> # OPERATORS
>>> ## Operations without multipliers
>>> ## Doubled Operators
>>> v.process_keys('~~'); v
sOMETHING CAME BEFOR[E
>>> v.process_keys('~~'); v
Something came befor[e
>>> v.process_keys('yy'); v.yank
'Something came before'
>>> v.process_keys('ccThis is a new sentence'+ESC); v
This is a new sentenc[e
>>> v.process_keys('zz'); v
_____________________[_
>>> v.yank
'This is a new sentence'
>>> v.process_keys('dd'); v
[
>>> ## Standard Operations
>>> v.process_keys('iNew test sentence has six words.'); v
New test sentence has six words.|
>>> v.process_key(ESC)
>>> ### tests of w motion
>>> v.process_keys('0w'); v
New [test sentence has six words.
>>> v.process_keys('d'); v.mode
'motion_pending'
>>> v.process_keys('w'); v
New [sentence has six words.
>>> v.process_keys('frdw'); v
New sentence has six w[o
>>> v.process_keys('u'); v
New [sentence has six words.
>>> v.process_keys('cwinner'); v
New inner| has six words.
>>> v.process_key(ESC)
>>> ### tests of operators with miscellaneous motions
>>> v.process_keys('y0'); v.yank
'New inner'
>>> v.process_keys('~$'); v
New inne[R HAS SIX WORDS.
>>> # TAB COMPLETION
>>> v.process_keys('dd'); v
[
>>> v.completions = ['this', 'that', 'theother']
>>> v.process_keys('ieither\t'); v
either|
>>> v.process_keys(' \t'); v
either |
>>> v.process_keys('t\t'); v
either this|
>>> v.process_keys('\t\t'); v
either theother|
>>> v.process_keys('\t'); v
either t|



'''
contains the VimEditor class and instantiates a single instance of it.
by using the same instance we can preserve yank and line_history for
the duration of the vinca_session.

'''
import re
import string

from vinca.lib import ansi
from vinca.lib import terminal
from vinca.lib.readkey import readkey, keys

# TODO dff cft &c. 
# TODO dot repeat
# TODO dot repeat insertions like cwnewtext
class VimEditor:
	# mutable variables like line_history and yank_history
	# will be shared across all instances of the class
	line_history = []
	yank_history = ['']
	green, red = ansi.codes['green'], ansi.codes['red']
	reset = ansi.codes['reset']
	mode_prompt_dict = {
	'normal':		      f'{red}[N]{reset} ',
	'motion_pending':             f'{red}[N]{reset} ',
	'search_char_pending':        f'{red}[N]{reset} ',
	'replace_char_pending':       f'{red}[N]{reset} ',
	'insert':       	      f'{green}[I]{reset} '}
	MODES = ('normal','insert','motion_pending',
		'search_char_pending', 'replace_char_pending')
	OPERATORS = 'dcy~z'
	MOTIONS = 'webftFT;,hl()0$_^'
	ACTIONS = 'sSru.xXDCpPYaiAIjk'
	DIGITS = '123456789'
	BOW = re.compile(	# BEGINNING OF WORD
		'((?<=\s)|^)'	# pattern is preceded by whitespace or BOL
		'\w')		# beginning of word is any alphanumeric.
	EOW = re.compile(	# END OF WORD
		'\w'		# any alphanumeric character
		'(\s|$)')	# succeeded by whitespace or EOL
	EOS = re.compile('[.!?]')  # EDGE OF SENTENCE

	@property
	def yank(self):
		return self.yank_history[-1]


	def __init__(self, text='', mode='default', prompt = '', completions = None):
		self.text = text
		self.mode = mode
		if mode=='default':
			if self.text:
				self.mode = 'normal'
			elif not self.text:
				self.mode = 'insert'
		self.prompt = prompt
		self.completions = completions if completions else []

		self.pos = 0
		self.completing = False

		self.multiplier = 1
		self.undo_stack = []
		self.redo_stack = []
		self.operator = None
		# store previous keypresses to redo actions with .
		self.current_insertion = ''
		self.prev_multiplier = 1
		self.prev_insertion = ''
		self.prev_action = None
		self.prev_operator = None
		self.prev_motion = None
		self.thing_to_repeat = 'none'
		# use up and down arrows to scroll through previous entries
		self.line_history_idx = 0  
		# character jumping (using f and t)
		self.search_char = ''
		self.char_jumper = None
		# delimit the range of motion for an operation
		self.start = None  
		self.end = None

	def process_key(self, key):
		if self.mode == 'insert':
			self.do_insert(key)
		elif self.mode == 'normal':
			self.process_key_normal_mode(key)
		elif self.mode == 'motion_pending':
			self.process_key_motion_pending_mode(key)
		elif self.mode == 'search_char_pending':
			self.process_key_search_char_pending_mode(key)
		elif self.mode == 'replace_char_pending':
			self.process_key_replace_char_pending_mode(key)

	def process_keys(self, keys, debug_id = 0):
		self.debug_id = debug_id
		for key in keys:
			self.process_key(key)

	def process_key_search_char_pending_mode(self, key):
		assert self.char_jumper in ('f','F','t','T')
		sc = self.search_char = key
		if sc in '[].?':
			sc = '\\' + sc
		self.pos = {'f': self.idx(sc),
			    'F': self.idx(sc, back = True),
			    't': self.idx(f'.(?={sc})'),
			    'T': self.idx(f'(?<={sc}).', back = True),
			   }[self.char_jumper]
		self.mode = 'normal'

	def process_key_replace_char_pending_mode(self, key):
		before = self.text[:self.pos]
		after  = self.text[self.pos+1:]
		self.text = before + key + after
		self.mode = 'normal'
		self.save_state()

	def process_key_motion_pending_mode(self, key):
		assert self.mode == 'motion_pending'
		assert self.operator
		# multiplier
		if key in self.DIGITS:
			self.multiplier = int(key)
			return
		# determine [start:end] range for the multiplier
		if key == self.operator:
			# allow for a doubled operator to act on the line
			# e.g. dd or yy is shorthand for
			# 0d$ and 0y$
			self.start = 0
			self.end = len(self.text) - 1
		elif key in self.MOTIONS:
			self.start = self.pos
			for _ in range(self.multiplier):
				self.do_motion(key)
			self.reset_multiplier()
			self.end = self.pos
			self.pos = self.start
		else:
			return  # invalid motion
		self.do_operation()
		self.thing_to_repeat = 'operation'
		if self.mode == 'motion_pending':
			self.mode = 'normal'
		if self.mode != 'insert':
			# save state after performing an operation
			# except if in insert mode,
			# because we will save the state after
			# performing the insertion.
			self.save_state()

	def process_key_normal_mode(self, key):
		assert self.mode == 'normal'
		if key in self.DIGITS:
			self.multiplier = int(key)
		elif key in self.ACTIONS:
			for _ in range(self.multiplier):
				self.do_action(key)
			self.reset_multiplier()
			self.thing_to_repeat = 'action'
			if self.mode != 'insert' and key != 'u':
				# save state after performing an action
				# except if in insert mode,
				# because we will save the state after
				# performing the insertion.
				self.save_state()
		elif key in self.OPERATORS:
			self.mode = 'motion_pending'
			self.operator = key
		elif key in self.MOTIONS:
			for _ in range(self.multiplier):
				self.do_motion(key)
			self.reset_multiplier()

	def reset_multiplier(self):
		self.prev_multiplier = self.multiplier
		self.multiplier = 1

	def idx(self, pattern, back = False):
		'''Return the index of the next match of the pattern in the text.
		If we find no match we return the current pos.'''
		# list of matching indicies
		indices = [m.start() for m in re.finditer(pattern, self.text)] 
		if not back:
			return min([i for i in indices if i > self.pos],
				default = self.pos)
		if back:
			return max([i for i in indices if i < self.pos],
				default = self.pos)

	def save_state(self):
		self.undo_stack.append({'text':self.text,'pos':self.pos})

	@property
	def current_char(self):
		if not self.text:
			return None
		if self.mode == 'normal':
			return self.text[self.pos]
		elif self.mode == 'insert':
			return self.text[self.pos - 1]

	@property
	def prev_char(self):
		if self.pos == 0:
			return None
		return self.text[self.pos - 1]

	def do_tab_complete(self):
		# we want to tab through various completions
		# get the starting index of the selected word
		if self.current_char not in string.ascii_letters:
			return
		if self.prev_char and self.prev_char not in string.ascii_letters:
			BOW_idx = self.pos
		else:
			BOW_idx = self.idx(self.BOW, back=True)
			
		# if we are pressing tab for the first time
		# we need to determine the completion stem
		# and make a list of possible completions
		if not self.completing:
			stem = self.text[BOW_idx:self.pos+1]
			cc = [stem]
			cc += [c for c in self.completions if c.startswith(stem)] 
			self.current_completions = cc
			self.completing = True
			self.complete_idx = 0
		# now we do the tab completions
		# go to next match
		self.complete_idx += 1
		# allow for scrolling through from the last match to the first match
		self.complete_idx = self.complete_idx % len(self.current_completions)
		
		before = self.text[:BOW_idx]
		after = self.text[self.pos+1:]
		completion = self.current_completions[self.complete_idx]
		self.text = before + completion + after
		self.pos = BOW_idx + len(completion)

	def do_insert(self, key):
		if key == keys.ESC:
			self.prev_insertion = self.current_insertion
			self.save_state()
			self.current_insertion = ''
			self.mode = 'normal'
			self.pos -= 1
			self.bring_position_into_text()
			return
		if key == '\t' and self.completions:
			self.do_tab_complete()
			return
		else:
			self.completing = False
		if key == '\x1b\t':
			key = '\t'
		if key == '\x1b\r':
			key = '\n'
		if key in (keys.LEFT, keys.RIGHT):
			if key == keys.LEFT:
				self.pos -= 1
			elif key == keys.RIGHT:
				self.pos += 1
			self.bring_position_into_text()
			self.current_insertion = ''
		elif key == keys.BACK and self.pos > 0:
			self.text = self.text[:self.pos - 1] + self.text[self.pos:]
			self.pos -= 1
			self.current_insertion = ''
		elif key in string.printable:
			self.text = self.text[:self.pos] + key + self.text[self.pos:]
			self.current_insertion += key
			self.pos += 1

	def bring_position_into_text(self):
		min_pos = 0
		max_pos = len(self.text) - 1 if self.text else 0
		if self.pos < min_pos:
			self.pos = min_pos
		elif self.pos > max_pos:
			self.pos = max_pos

	def do_operation(self):
		self.start, self.end = min(self.start, self.end), max(self.start, self.end)
		key = self.operator
		if key == 'd':
			self.text = self.text[:self.start] + self.text[self.end+1:]
			self.bring_position_into_text()
		if key == 'c':
			self.text = self.text[:self.start] + self.text[self.end+1:]
			self.mode = 'insert'
		if key == 'y':
			self.yank_history.append(self.text[self.start:self.end+1])
		if key == '~':
			self.text = self.text[:self.start] + \
				self.text[self.start:self.end+1].swapcase() + \
				self.text[self.end+1:]
		if key == 'z':
			self.yank_history.append(self.text[self.start:self.end+1])
			self.text = self.text[:self.start] + \
				'_'*(self.end-self.start+1) + self.text[self.end+1:]
		self.prev_operator = key

	def w_motion(self):
		if self.mode == 'normal':
			return self.idx(self.BOW)
		elif self.mode == 'motion_pending':
			if self.operator in 'c~yz':
				return self.idx(self.EOW)
			elif self.operator == 'd':
				# dw should delete up through the
				# beginning of the next word ( \w)
				# or all the way thru to the end of the line
				pat = re.compile('(( \w)|$)')
				return self.idx(pat)

	def do_motion(self, key):
		# jump to character
		if key in ('f','F','t','T'):
			self.char_jumper = key
			self.mode = 'search_char_pending'
			return
		# other motions
		self.pos = {
			# jump by word
			'w': self.w_motion,
			'e': lambda: self.idx(self.EOW),
			'b': lambda: self.idx(self.BOW, back = True),
			# repeat character jumps
			';': lambda: self.idx(self.search_char),
			',': lambda: self.idx(self.search_char, back = True),
			# left / right navigation
			'h':       lambda: max(0, self.pos-1),
			'l': lambda: min(len(self.text)-1, self.pos+1),
			# sentence jumping
			')': lambda: self.idx(self.EOS),
			'(': lambda: self.idx(self.EOS, back = True),
			# jump to beginning or end of line
			'0': lambda: 0,
			'^': lambda: 0,
			'_': lambda: 0,
			'$': lambda: len(self.text)
			}[key]()
		self.prev_motion = key

	def do_action(self, key): 
		k = key
		# substitution
		if k == 'C':
			self.process_keys('c$')
		if k in ('s','S'):
			self.mode = 'insert'
			if key == 'S':
				self.text = ''
			if key == 's':
				self.text = self.text[:self.pos] + self.text[self.pos+1:]
		if k == 'r':
			self.mode = 'replace_char_pending'
		# reversion and redoing
		if k == 'u' and self.undo_stack:
			# save current state
			current_state = self.undo_stack.pop()
			self.redo_stack.append(current_state)
			# retrieve previous state
			prev_state = self.undo_stack[-1]
			self.text, self.pos = prev_state['text'], prev_state['pos']
		if k == keys.CTRL_R and self.redo_stack:
			self.undo_stack.append({'text': self.text, 'pos': self.pos})
			new_state = self.redo_stack.pop()
			self.text, self.pos = new_state['text'], new_state['pos']
		if k == '.':
			self.repeat()
		# deletion
		if k == 'D':
			self.process_keys('d$')
		if k in ('x','X'):
			self.text = {
				'x': self.text[:self.pos] + self.text[self.pos+1:],
				'X': self.text[:self.pos-1] + self.text[self.pos:]}[k]
			self.pos -= (k == 'X')
		# copy and paste
		if k == 'Y':
			self.process_keys('y$')
		if k == 'p':
			self.text = self.text[:self.pos+1] + \
				    self.yank + self.text[self.pos+1:]
		if k == 'P':
			self.text = self.text[:self.pos] + \
			 	    self.yank + self.text[self.pos:]
			self.pos += len(self.yank)
		# enter insert mode
		if k in ('i','I','a','A'):
			self.mode = 'insert'
			self.pos =  {'i': self.pos,
				'I': 0,
				'a': self.pos + 1,
				'A': len(self.text)}[k]
				
		# history scrolling
		if k in ('j','k',keys.DOWN, keys.UP):
			max_lhi = 0
			min_lhi = - len(self.line_history)
			if k in ('j', keys.DOWN):
				if self.line_history_idx < max_lhi:
					self.line_history_idx += 1
			if k in ('k', keys.UP):
				if self.line_history_idx > min_lhi:
					self.line_history_idx -= 1
			lhi = self.line_history_idx
			if lhi == 0:
				self.text = ''
			elif lhi <= -len(self.text):
				self.text = self.line_history[0]
			else:
				self.text = self.line_history[lhi]
			self.pos = len(self.text) - 1

		self.prev_action = key if key != '.' else self.prev_action

	def repeat(self):
		if self.thing_to_repeat == 'none':
			return
		if self.thing_to_repeat == 'action':
			assert self.prev_action
			for _ in range(self.prev_multiplier):
				self.do_action(self.prev_action)
				if self.mode == 'insert':
					self.text = self.text[:self.pos] + \
						self.prev_insertion + \
						self.text[self.pos:]
					self.mode == 'normal'
		if self.thing_to_repeat == 'operation':
			assert self.prev_operator and self.prev_motion
			for _ in range(self.prev_multiplier):
				self.do_operation(self.prev_operator)

	def __str__(self):
		mode_prompt = self.mode_prompt_dict[self.mode]
		prefix = mode_prompt + self.prompt
		cursor = ansi.codes['reverse'] + \
			(self.text[self.pos] if self.pos<len(self.text) else ' ') + \
			ansi.codes['reset']
		text_with_cursor = self.text[:self.pos] + cursor + self.text[self.pos+1:]
		return prefix + text_with_cursor

	def __repr__(self):
		cursor = '|' if self.mode == 'insert' else '['
		text_with_cursor = self.text[:self.pos] + cursor + self.text[self.pos:]
		return text_with_cursor
		

	def scrollback(self):
		l = terminal.count_screen_lines(str(self))
		ansi.up_line(l)
		ansi.clear_to_end()

	def run(self):
		with terminal.NoCursor():
			while True:
				print(self)
				key = readkey()
				if key in ('\n','\r'):
					break
				self.scrollback() # clear for next iter
				self.process_key(key)	

			self.line_history.append(self.text)
			return self.text

if __name__ == '__main__':
	import doctest
	test_vim = VimEditor(text = 'The quick brown fox jumped over the lazy dog.')
	doctest.testfile('vim_tests.py',
			optionflags=doctest.REPORT_ONLY_FIRST_FAILURE,
			globs = {'v': test_vim, 'ESC':keys.ESC, 'BACK':keys.BACK,
				 'LEFT':keys.LEFT, 'RIGHT':keys.RIGHT})



import importlib
from importlib import import_module

def schedule(name, history):

	# import the specific scheduler module
	m = importlib.import_module('.' + name, package = 'vinca.schedulers')
	# invoke the specific scheduler
	return m.schedule(history)



from pathlib import Path
import vinca.card
from vinca.config import config

vinca_path = Path(__file__).parent
tags_path = vinca_path / 'tags'

class TagsCache(list):
	def __init__(self):
		tags = tags_path.read_text().splitlines()
		super().__init__(tags)

	def update(self):
		'rebuild the tags list'
		# TODO a collector module
		all_cards = [vinca.card.Card(p.name) for p in config.cards_path.iterdir()] 
		self[:] = [tag for card in all_cards for tag in card.tags]
		self.save()

	def save(self):
		tags_path.write_text('\n'.join(self))
	
	def add_tag(self, tag):
		if tag not in self:
			self.append(tag)
			self.save()

	def add_tags(self, tags):
		for tag in tags:
			self.add_tag(tag)

tags_cache = TagsCache()

	
		



'''Vinca 112
Simple Spaced Repetition'''

import inspect as _inspect
from pathlib import Path as _Path
from vinca.cardlist import Cardlist as _Cardlist
from vinca.card import Card as _Card
from vinca.config import config as _config
from vinca.generators import generators_dict as _generators_dict
import fire as _fire

# load the card generators into the module
for _hotkey, _generator_func in _generators_dict.items():
	locals()[_generator_func.__name__] = _generator_func
	locals()[_hotkey] = _generator_func


# create a collection (cardlist) out of all the cards
_ALL_CARDS = [_Card(p) for p in _config.cards_path.iterdir()] 
col = collection = _Cardlist(_ALL_CARDS)
# import all the methods of the collection object directly into the module's namespace
# this is so that ```vinca col filter``` can be written more shortly as ```vinca filter```
for _method_name, _method in _inspect.getmembers(col):
	locals()[_method_name] = _method

# utility functions
help = '''\
vinca --help           general help
vinca filter --help    help on a specific subcommand
man vinca              vinca tutorial'''
version = lambda: 113

# quick reference to the most recent card
_lcp = _config.last_card_path
_lcp_exists = isinstance(_lcp, _Path) and _lcp.exists()
lc = last_card = _Card(path = _lcp) if _lcp_exists else 'no last card'

# move config.set_cards_path into locals
set_cards_path = _config.set_cards_path
cards_path = _config.cards_path

_fire.Fire()
'''
Add the following code to the ActionGroup object in helptext.py of fire to get proper aliasing
A better way would be to go back further into the code and check if two functions share the same id

  def Add(self, name, member=None):
    if member and member in self.members:
      dupe = self.members.index(member)
      self.names[dupe] += ', ' + name
      return
    self.names.append(name)
    self.members.append(member)
'''
'''
Make this substitution on line 458 of core.py to allow other iterables to be accessed by index

    # is_sequence = isinstance(component, (list, tuple))
    is_sequence = hasattr(component, '__getitem__') and not hasattr(component, 'values')
'''
'''
And make a corresponding change in generating the help message

  is_sequence = hasattr(component, '__getitem__') and not hasattr(component, values)
  # if isinstance(component, (list, tuple)) and component:
  if is_sequence and component:
'''



import subprocess
import json
import datetime
TODAY = datetime.date.today()
DAY = datetime.timedelta(days=1)
from pathlib import Path
from shutil import copytree

from vinca import reviewers, editors, schedulers 
from vinca.tag_caching import tags_cache
from vinca.config import config
from vinca.lib.vinput import VimEditor
from vinca.lib.random_id import random_id
from vinca.history import History, HistoryEntry

class Card:
	# Card class can load without 
	default_metadata = {'editor': 'base', 'reviewer':'base', 'scheduler':'base',
			    'tags': [], 'history': History([HistoryEntry(TODAY, 0, 'create')]), 'deleted': False,
			    'due_date': TODAY, 'string': ''}

	def __init__(self, path=None, create=False):
		assert (path is not None) ^ create
		if path:
			self.init_loaded_card(path)
		elif create:
			self.init_new_card()
		self._hotkeys = {'e': self.edit,
				'E': self.edit_metadata,
				'M': self.print_metadata,
				't': self.edit_tags,
				'd': self.delete,
				's': self.summarize,
				'r': self.review,
				'+': self.postpone}
		self._confirm_exit_commands = [self.print_metadata, self.summarize]

	def init_loaded_card(self, path):
		self.path = path
		self.metadata_is_loaded = False

	def init_new_card(self):
		# create card location
		self.path = self.new_path()
		self.path.mkdir()
		# initialize metadata
		self.metadata = Card.default_metadata
		self.metadata_is_loaded = True
		self.save_metadata()
		# easy access to the last created card
		config.last_card_path = self.path

	def summarize(self):
		s  = f'due	{self.due_date}\n'
		s += f'seen	{len(self.history)}x\n'
		s += f'time	{self.history.time}s\n'
		print(s)

	@property
	def metadata_path(self):
		return self.path / 'metadata.json'

	def load_metadata(self):
		self.metadata = json.load(self.metadata_path.open())
		# dates must be serialized into strings for json
		# I unpack them here
		self.metadata['history'] = History.from_json_list(self.metadata['history'])
		assert self.metadata['history'], f'empty history metadata for {self.path}'
		self.metadata['due_date'] = datetime.date.fromisoformat(self.metadata['due_date'])
		self.metadata_is_loaded = True

	def save_metadata(self):
		json.dump(self.metadata, self.metadata_path.open('w'), default=str, indent=2)

	for m in default_metadata.keys():
		# create getter and setter methods for everything in the metadata dictionary
		exec(f'''
@property
def {m}(self):
	if not self.metadata_is_loaded:
		self.load_metadata()
	return self.metadata["{m}"]''')
		exec(f'''
@{m}.setter
def {m}(self, new_val):
	if not self.metadata_is_loaded:
		self.load_metadata()
	self.metadata["{m}"] = new_val
	self.save_metadata()''')	

	# overwrite the tags setter with one modification
	# we want to update the tags_cache
	@tags.setter
	def tags(self, tags):
		if not self.metadata_is_loaded:
			self.load_metadata()
		self.metadata['tags'] = tags
		self.save_metadata()
		tags_cache.add_tags(tags)

	def __str__(self):
		return self.string

	def review(self):
		reviewers.review(self)
		self.save_metadata() # we have probably appended a history entry
		self.schedule()

	def make_string(self):
		self.string = reviewers.make_string(self)

	def edit(self):
		editors.edit(self) 
		self.make_string()
		self.save_metadata() # we have probably modified history

	def edit_metadata(self):
		subprocess.run(['vim',self.path/'metadata.json'])
		self.load_metadata()
		self.make_string()

	def print_metadata(self):
		for k,v in self.metadata.items():
			history_newline = '\n' if k=='history' else ''
			print(f'{k:20}', history_newline, v, sep='', end='\n\n')

	def schedule(self):
		if dd := schedulers.schedule(name=self.scheduler, history=self.history):
			self.due_date = dd

	def copy(self, new_path):
		new_path.mkdir()
		copytree(self.path, new_path)

	def new_path(self):
		return config.cards_path / ('card-' + random_id())

	def delete(self, toggle=True):
		if toggle:
			self.deleted = not self.deleted
		elif not toggle:
			self.deleted = True

	def due_as_of(self, date):
		return self.due_date <= date

	@property
	def is_due(self):
		return self.due_as_of(TODAY)

	def edit_tags(self):
		self.tags = VimEditor(prompt = 'tags: ', text = ' '.join(self.tags), completions = tags_cache).run().split()

	def postpone(self):
		tomorrow = TODAY + DAY
		self.due_date = tomorrow

	@property
	def create_date(self):
		return self.history.create_date

	@property
	def seen_date(self):
		return self.history.last_date

	@property
	def new(self):
		return self.history.new


	@property
	def time(self):
		return self.history.time



import datetime
DAY = datetime.timedelta(days=1)

def schedule(history):
	'''
	>>> import sys
	>>> sys.path.append('/home/oscar/vinca/vinca')
	>>> from history import test1, test2, test3, test4, test5
	>>> test1
	DATE          TIME    GRADE
	2021-11-07       0    create
	>>> schedule(test1)
	>>> test2
	DATE          TIME    GRADE
	2021-11-01       0    create
	2021-11-01      99    edit
	2021-11-01      11    exit
	2021-11-01     133    edit
	2021-11-01       9    good
	2021-11-02       9    good
	2021-11-05       2    exit
	2021-11-05       9    good
	>>> schedule(test2)
	datetime.date(2021, 11, 12)
	>>> test3
	DATE          TIME    GRADE
	2021-11-01       0    create
	2021-11-01      99    edit
	2021-11-05       9    hard
	>>> schedule(test3)
	datetime.date(2021, 11, 7)
	>>> test4
	DATE          TIME    GRADE
	2021-11-01       0    create
	2021-11-01      11    exit
	2021-11-01       9    good
	2021-11-05       9    again
	>>> schedule(test4)
	datetime.date(2021, 11, 5)
	>>> test5
	DATE          TIME    GRADE
	2021-11-01       0    create
	2021-11-01      11    edit
	2021-11-01       9    good
	2021-11-05       9    easy
	>>> schedule(test5)
	datetime.date(2021, 11, 18)
	'''


	if history.last_grade in history.admin_grades:
		return
	if history.last_grade == 'again':
		return history.last_date 
	if history.last_grade == 'hard':
		return history.last_date + DAY * (history.last_interval / 2).days
	if history.last_grade == 'good':
		return history.last_date + (history.last_interval * 2) + DAY
	if history.last_grade == 'easy':
		return history.last_date + (history.last_interval * 3) + DAY

if __name__ == '__main__':
	import doctest
	doctest.testmod()



from vinca.lib import ansi
import re
import shutil

COLUMNS, LINES = shutil.get_terminal_size()

# A Context Manager for the terminal's alternate screen
class AlternateScreen:
	def __enter__(self):
		ansi.save_cursor()
		ansi.hide_cursor()
		ansi.save_screen()
		ansi.clear()
		ansi.move_to_top()

	def __exit__(self, *exception_args):
		ansi.restore_screen()
		ansi.restore_cursor()


class NoCursor:
	def __enter__(self):
		ansi.hide_cursor()

	def __exit__(self, *exception_args):
		ansi.show_cursor()
		

class LineWrapOff:
	def __enter__(self):
		ansi.line_wrap_off()

	def __exit__(self, *exception_args):
		ansi.line_wrap_on()

def roundup(flt):
	'''
	>>> roundup(5)
	5
	>>> roundup(4.01)
	5
	>>> roundup(5.000)
	5
	'''
	if int(flt) == flt:
		return int(flt)
	return round(flt + 0.5)

def count_screen_lines(text):
	'''
	>>> COLUMNS
	20
	>>> count_screen_lines('the quick fox')
	1
	>>> count_screen_lines('-'*20)
	1
	>>> count_screen_lines('this very long text' * 100)
	95
	>>> count_screen_lines('long'*100 + '\\ris overwritten')
	1
	>>> count_screen_lines('\033[?25l hidden cursor \033[?25h')
	1
	'''
	text = strip_ansi(text)
	lines = text.split('\n')
	# carriage return overwrites the line so we want
	# the text after the last carriage return on the line
	lines = [l.split('\r')[-1] for l in lines] 
	lines = [roundup(len(l) / COLUMNS) for l in lines]
	return sum(lines)

def strip_ansi(text):
	'''
	>>> strip_ansi('the quick fox')
	'the quick fox'
	>>> strip_ansi('\033[1mBOLD')
	'BOLD'
	>>> strip_ansi('\033[?25l hidden cursor \033[?25h')
	' hidden cursor '
	>>> strip_ansi('down\033E line')
	'down line'
	'''
	ansi_escape = re.compile(r'''
	    \x1B  # ESC
	    (?:   # 7-bit C1 Fe (except CSI)
		[@-Z\\-_]
	    |     # or [ for CSI, followed by a control sequence
		\[
		[0-?]*  # Parameter bytes
		[ -/]*  # Intermediate bytes
		[@-~]   # Final byte
	    )
	''', re.VERBOSE)
	result = ansi_escape.sub('', text)
	return result

if __name__ == '__main__':
	import doctest
	COLUMNS = 20
	doctest.testmod(extraglobs = {'COLUMNS': 20})



from vinca.lib import ansi
from vinca.lib.terminal import LineWrapOff, AlternateScreen
from vinca.lib.readkey import readkey, keys
from vinca.generators import generators_dict

FRAME_WIDTH = 10

class Browser():
	quit_keys = ('q','\n','\r',keys.ESC)
	move_keys = ('j','k',keys.DOWN, keys.UP)

	def __init__(self, cardlist):
		self.cardlist = cardlist
		self.reviewing = False
		self.sel = 0
		self.frame = 0

	@property
	def N(self):
		return len(self.cardlist)

	@property
	def selected_card(self):
		return self.cardlist[self.sel]

	@property
	def visible_lines(self):
		return min(self.N, FRAME_WIDTH) + (self.status_bar is not None)

	
	@property
	def status_bar(self):
		if not self.N > FRAME_WIDTH:
			return None
		return ansi.codes['light'] + f'{self.sel} of {self.N}' + ansi.codes['reset']


	def draw_browser(self):
		ansi.hide_cursor()
		with LineWrapOff():
			if self.status_bar:
				print(self.status_bar)
			for i, card in enumerate(self.cardlist[self.frame:self.frame+FRAME_WIDTH], start=self.frame):
				if card.is_due:
					ansi.bold()
					ansi.blue()
				if card.deleted:
					ansi.crossout()
					ansi.red()
				if i==self.sel:
					ansi.highlight()
				print(card)
				ansi.reset()

	def clear_browser(self):
		ansi.up_line(self.visible_lines)
		ansi.clear_to_end()

	def close_browser(self):
		self.clear_browser()
		ansi.show_cursor()
		exit()

	def redraw_browser(self):
		self.clear_browser()
		self.draw_browser()
		
	def move(self, key):
		if key in ('j',keys.DOWN) and self.sel < self.N - 1:
			self.move_down()
		if key in ('k',keys.UP) and self.sel > 0:
			self.move_up()

	def move_down(self):
		self.sel += 1
		# scroll down if we are off the screen
		self.frame += (self.frame + FRAME_WIDTH == self.sel)  
	
	def move_up(self):
		self.sel -= 1
		# scroll down if we are off the screen
		self.frame -= (self.frame - 1 == self.sel)  

	def review(self):
		self.reviewing = True
		self.browse()
	
	def browse(self):
		if not self.cardlist:
			print('no cards')
			return
		self.draw_browser()
		while True:

			# if we are in reviewing mode continue on to the next due card
			if self.reviewing:
				if self.selected_card.is_due:
					self.selected_card.review()
					self.selected_card.schedule()
					if self.selected_card.history.last_grade == 'exit':
						self.reviewing = False
				while not self.selected_card.is_due:
					self.move_down()
					if self.sel >= self.N - 1:
						self.close_browser()
				self.redraw_browser()
				continue

			if not self.reviewing:
				k = readkey()
				if k == 'R':
					self.reviewing = True

			if k in self.quit_keys:
				self.close_browser()

			if k in self.move_keys:
				self.move(k)

				
			if command := (self.cardlist._hotkeys.get(k) or self.selected_card._hotkeys.get(k)):
				with AlternateScreen():
					command()
					if command in self.selected_card._confirm_exit_commands + self.cardlist._confirm_exit_commands:
						print('press any key to continue')
						readkey()

			if generator := generators_dict.get(k):
				with AlternateScreen():
					new_card = generator()
					self.cardlist._insert_card(self.sel, new_card)
				ansi.down_line()
				if self.N == FRAME_WIDTH + 1:
					ansi.down_line()

			self.redraw_browser()







from vinca.card import Card

def two_lines():
	''' make a simple question and answer card '''
	new_card = Card(create=True)
	new_card.editor, new_card.reviewer, new_card.scheduler = 'two_lines', 'two_lines', 'base'
	(new_card.path / 'front').touch()
	(new_card.path / 'back').touch()
	new_card.edit()
	return new_card



import re
import datetime
from shutil import copytree, rmtree
from vinca. browser import Browser
from vinca.lib import ansi
from vinca.lib.vinput import VimEditor
from vinca.lib.readkey import readkey
from vinca.lib import casting
from vinca.config import config
from vinca.tag_caching import tags_cache
TODAY = datetime.date.today()

class Cardlist:
	''' this is a collection of cards. most of the user interface takes place through the browser '''


	def __init__(self, _cards):
		# we do not subclass list so that
		# the fire help is not littered
		# with inherited methods
		self._cards = _cards
		self._hotkeys = {'D': self.delete,
				 'T': self.edit_tags,
				 'C': self.count}
		self._confirm_exit_commands = [self.count]

	def __iter__(self):
		return iter(self._cards)

	def __len__(self):
		return len(self._cards)

	def __getitem__(self, slice):
		return self._cards[slice]

	def _insert_card(self, idx, card):
		self._cards.insert(idx, card)

	def __str__(self):
		s = ''
		l = len(self)
		if l == 0:
			return 'No cards.'
		if l > 10:
			s += f'10 of {l}\n'
		s += ansi.codes['line_wrap_off']
		for card in self[:10]:
			if card.due_as_of(TODAY):
				s += ansi.codes['bold']
				s += ansi.codes['blue']
			if card.deleted:
				s += ansi.codes['crossout']
				s += ansi.codes['red']
			s += f'{card}\n'
			s += ansi.codes['reset']
		s += ansi.codes['line_wrap_on']
		return s

	def browse(self):
		''' scroll through your collection with j and k '''
		Browser(self).browse()
	b = browse

	def review(self):
		''' review all cards '''
		Browser(self).review()
	r = review
				
	def add_tag(self, tag):
		for card in self:
			card.tags += [tag]

	def remove_tag(self, tag):
		for card in self:
			if tag in card.tags:
				card.tags.remove(tag)
			# TODO do this with set removal
			card.save_metadata()  # metadata ops should be internal TODO

	def count(self):
		''' simple summary statistics '''
		total_count = len(self)
		due_count = len(self.filter(due_only=True))
		print(f'total	{total_count}')
		print(f'due	{due_count}')

	def edit_tags(self):
		''' add or remove tags from all cards '''
		tags_add = VimEditor(prompt = 'tags to add: ', completions = tags_cache).run().split()
		tags_remove = VimEditor(prompt = 'tags to remove: ', completions = tags_cache).run().split()
		for tag in tags_add:
			self.add_tag(tag)
		for tag in tags_remove:
			self.remove_tag(tag)

	def save(self, save_path):
		''' backup your cards '''
		save_path = casting.to_path(save_path)
		for card in self:
			copytree(card.path, save_path / card.path.name)

	@staticmethod
	def load(load_path, overwrite = False):
		''' load cards from another folder into your collection '''
		load_path = casting.to_path(load_path)
		if overwrite:
			print(f'Overwrite {len(self)} cards? (y/n)')
			if readkey == 'y':
				rmtree(config.cards_path)
				copytree(load_path, config.cards_path)
			return
		for card_path in load_path.iterdir():
			copytree(card_path, config.cards_path / card_path.name)


	def purge(self):
		''' Permanently delete all cards marked for deletion. '''
		deleted_cards = self.filter(deleted_only = True)
		if not deleted_cards:
			print('no cards are marked for deletion.')
			return
		print(f'delete {len(deleted_cards)} cards? (y/n)')
		if readkey() == 'y':
			for card in deleted_cards:
				rmtree(card.path)

	def delete(self):
		for card in self:
			card.delete(toggle = True)



	def filter(self, *,
		   tag = None,
		   create_date_min=None, create_date_max=None,
		   seen_date_min=None, seen_date_max=None,
		   due_date_min=None, due_date_max=None,
		   editor=None, reviewer=None, scheduler=None,
		   deleted_only=False, 
		   due_only=False,
		   new_only=False,
		   invert=False):
		''' filter the collection by a wide array of predicates '''
		
		# cast dates to dates
		create_date_min = casting.to_date(create_date_min)
		create_date_max = casting.to_date(create_date_max)
		seen_date_min = casting.to_date(seen_date_min)
		seen_date_max = casting.to_date(seen_date_max)
		due_date_min = casting.to_date(due_date_min)
		due_date_max = casting.to_date(due_date_max)

		if due_only: due_date_max = TODAY
		# compile the regex pattern for faster searching

		f = lambda card: (((not tag or tag in card.tags) and
				(not create_date_min or create_date_min <= card.create_date) and
				(not create_date_max or create_date_max >= card.create_date) and 
				(not seen_date_min or seen_date_min <= card.seen_date) and
				(not seen_date_max or seen_date_max >= card.seen_date) and 
				(not due_date_min or due_date_min <= card.due_date) and
				(not due_date_max or due_date_max >= card.due_date) and 
				(not editor or editor == card.editor) and
				(not reviewer or reviewer == card.reviewer) and
				(not scheduler or scheduler == card.scheduler) and
				(not deleted_only or card.deleted ) and
				(not new_only or card.new)) ^
				invert)
		
		# matches.sort(key=lambda card: card.seen_date, reverse=True)
		return self.__class__([c for c in self if f(c)])
	f = filter

	def find(self, pattern):
		''' return the first card containing a search pattern '''
		matches = self.findall(pattern)
		matches.sort(criterion = 'seen-date')
		return matches[0] if matches else 'no match found'

	def findall(self, pattern):
		''' return all cards containing a search-pattern '''
		p = re.compile(f'({pattern})')  # wrap in parens to create regex group \1
		contains_pattern = lambda card: p.search(card.string)
		return self.__class__([c for c in self if contains_pattern(c)])

	def sort(self, criterion, *, reverse=False):
		''' sort the collection. criterion
		should be (due-date | seen-date | create-date | time) '''
		crit_dict = {
			'due-date': lambda card: card.create_date,
			'seen-date': lambda card: card.seen_date, 
			'create-date': lambda card: card.create_date,
			'time': lambda card: card.time}
		reverse_crits = ('create-date', 'seen-date', 'time')
		if criterion not in crit_dict:
			print('supply a criterion: create_date | seen_date | due_date')
		reverse = reverse ^ (criterion in reverse_crits)
		self._cards.sort(key = crit_dict[criterion], reverse = reverse)
		return self
	s = sort

	def time(self):
		''' return the total time spend studying these cards '''
		return sum([card.history.time for card in self], start=datetime.timedelta(0))

	def range(self):
		''' useful for writing shell functions '''
		for i, card in enumerate(self):
			print(i)

	def cut(self, idx):
		self._cards[idx:] = []
		return self
	




from pathlib import Path
import datetime

TODAY = datetime.date.today()
DAY = datetime.timedelta(days=1)

def to_path(arg):
	''' convert a string to a path and verify it exists
	tests
	>>> to_path('/home')
	PosixPath('/home')
	>>> to_path(Path('/home'))
	PosixPath('/home')
	>>> to_path('/nowhere')
	Traceback (most recent call last):
	...
	FileNotFoundError
	'''
	if type(arg) is Path:
		return arg
	p = Path(arg)
	if not p.exists():
		raise FileNotFoundError(p)
	return p

def to_date(arg):
	''' convert 2021-10-31 to a October 31, 2021
	and convert -7 to one week before today
	>>> to_date('2021-10-31')
	datetime.date(2021, 10, 31)
	>>> TODAY - to_date(-7)
	datetime.timedelta(days=7)
	>>> to_date('10-31-2021')
	Traceback (most recent call last):
	...
	ValueError: Dates must be YYYY-MM-DD
	'''
	if type(arg) is datetime.date:
		return arg
	if type(arg) is str:
		try:
			return datetime.date.fromisoformat(arg)
		except ValueError:
			raise ValueError('Dates must be YYYY-MM-DD')
	if type(arg) is int:
		return TODAY + arg*DAY  # date relative to today


if __name__ == '__main__':
	import doctest
	doctest.testmod()



import datetime
from collections import namedtuple
to_date = datetime.date.fromisoformat

HistoryEntry = namedtuple('HistoryEntry', ['date','time','grade'])

class History(list):
	admin_grades = ('create', 'edit','exit','preview')

	@classmethod
	def from_json_list(cls, entries):
		'''
		>>> History.from_json_list([['2021-11-07',0,'create']])
		DATE          TIME    GRADE
		2021-11-07       0    create
		'''
		return cls([HistoryEntry(to_date(d), t, g) for d, t, g in entries])

	@property
	def create_date(self):
		'''
		>>> test2.create_date
		datetime.date(2021, 11, 1)
		'''
		assert self[0].grade == 'create'
		return self[0].date

	@property
	def time(self):
		'''
		>>> test2.time
		272
		'''
		return datetime.timedelta(seconds = sum([entry.time for entry in self]))

	@property
	def pure(self):
		'''
		>>> test2.pure
		DATE          TIME    GRADE
		2021-11-01       9    good
		2021-11-02       9    good
		2021-11-05       9    good
		'''
		# exclude administrative entries; only keep actual reviewing grades
		return History([entry for entry in self if entry.grade not in self.admin_grades])

	@property
	def last_interval(self):
		'''
		>>> test1.last_interval
		datetime.timedelta(0)
		>>> test3.last_interval
		datetime.timedelta(days=4)
		>>> test2.last_interval
		datetime.timedelta(days=3)
		'''
		pure = self.pure
		if len(pure) == 0:
			return datetime.timedelta(days=0)
		if len(pure) == 1:
			return pure[-1].date - self.create_date
		return pure[-1].date - pure[-2].date


	@property
	def last_grade(self):
		'''
		>>> test2.last_grade
		'good'
		'''
		return self[-1].grade

	@property
	def last_date(self):
		'''
		>>> test2.last_date
		datetime.date(2021, 11, 5)
		'''
		return self[-1].date

	def append_entry(self, date, time, grade):
		'''
		>>> test1
		DATE          TIME    GRADE
		2021-11-07       0    create
		>>> test1.append_entry(datetime.date(2021, 11, 1), 99, 'edit')
		>>> test1
		DATE          TIME    GRADE
		2021-11-07       0    create
		2021-11-01      99    edit
		'''
		assert type(date) is datetime.date	
		NewEntry = HistoryEntry(date, time, grade)
		self.append(NewEntry)

	@property
	def new(self):
		'''
		>>> test1.new
		True
		'''
		return len(self.pure) == 0

	def __repr__(self):
		s =  'DATE          TIME    GRADE\n'
		s += '\n'.join([f'{d}{t:8d}    {g}' for d, t, g in self])
		return s
	__str__ = __repr__

test1 = History([HistoryEntry('2021-11-07',0,'create')])
test2 = History([
	HistoryEntry(date=datetime.date(2021, 11, 1), time=0, grade='create'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=99, grade='edit'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=11, grade='exit'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=133, grade='edit'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=9, grade='good'),
	HistoryEntry(date=datetime.date(2021, 11, 2), time=9, grade='good'),
	HistoryEntry(date=datetime.date(2021, 11, 5), time=2, grade='exit'),
	HistoryEntry(date=datetime.date(2021, 11, 5), time=9, grade='good')])
test3 = History([
	HistoryEntry(date=datetime.date(2021, 11, 1), time=0, grade='create'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=99, grade='edit'),
	HistoryEntry(date=datetime.date(2021, 11, 5), time=9, grade='hard')])
test4 = History([
	HistoryEntry(date=datetime.date(2021, 11, 1), time=0, grade='create'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=11, grade='exit'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=9, grade='good'),
	HistoryEntry(date=datetime.date(2021, 11, 5), time=9, grade='again')])
test5 = History([
	HistoryEntry(date=datetime.date(2021, 11, 1), time=0, grade='create'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=11, grade='edit'),
	HistoryEntry(date=datetime.date(2021, 11, 1), time=9, grade='good'),
	HistoryEntry(date=datetime.date(2021, 11, 5), time=9, grade='easy')])

if __name__ == '__main__':
	import doctest
	doctest.testmod()





import json
from pathlib import Path

# so far the config object only stores the path
# in which to look for cards
class Config(dict):
	path = Path(__file__).parent / 'config.json'
	
	def __init__(self):
		dict.__init__(self, self.load())
	
	def load(self):
		with open(self.path) as f:
			return json.load(f)

	def save(self):
		with open(self.path,'w') as f:
			json.dump(self, f)

	def set_cards_path(self, path):
		''' vinca will look for cards in the chosen directory '''
		self['cards_path'] = str(path)
		self.save()

	@property
	def last_card_path(self):
		return Path(self['last_card_path'])

	@last_card_path.setter
	def last_card_path(self, card_path):
		self['last_card_path'] = str(card_path)
		self.save()

	@property
	def cards_path(self):
		return Path(self['cards_path'])

config = Config()



import subprocess
from pathlib import Path
import shutil
from vinca.lib.vinput import VimEditor

vinca_path = Path(__file__).parent.parent.parent

path = Path(__file__).parent

# TODO rewrite without vim
def edit(card):
	front_path, back_path = (card.path/'front'), (card.path/'back')
	front = front_path.read_text()
	back = back_path.read_text()
	new_front = VimEditor(text = front, prompt = 'Q: ').run()
	front_path.write_text(new_front)
	new_back = VimEditor(text = back, prompt = 'A: ').run()
	for media in ('image_front', 'image_back', 'audio_front', 'audio_back'):
		media_file = card.path / media
		ref = VimEditor(prompt = f'select a file for {media}').run()
		if ref == 'delete' and media_file.exists():
			media_file.unlink()
		elif ref:
			source_file = Path.cwd() / ref
			assert source_file.exists(), f'{source_file} not found'
			shutil.copy(ref, media_file)




# A simple module for reading a keystroke from the user
# adapted from the getchar function in the click library

import tty
import termios
import sys
import contextlib
import os
from types import SimpleNamespace

keys = SimpleNamespace(ESC = '\x1b', UP = '\x1b[A', DOWN = '\x1b[B', RIGHT = '\x1b[C',
			LEFT = '\x1b[D', BACK = '\x7f', CTRL_R = '\x12')

@contextlib.contextmanager
def raw_terminal():
	fd = sys.stdin.fileno()

	old_settings = termios.tcgetattr(fd) # save state of terminal

	try:
		# the tty has two modes:
		# 'cooked' waits for the user to press enter
		# 'raw' passes the typed characters immediately
		tty.setraw(fd)
		yield fd
	finally:
		# restore state of terminal
		termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
		sys.stdout.flush() 


def readkey():
	with raw_terminal() as fd:
		# a single key can be represented by up to 4 bytes
		return os.read(fd, 4).decode()



import random
import string

def random_id():
	'''
	>>> len(random_id())
	8
	>>> type(random_id())
	<class 'str'>
	>>> 'AAAAAAAA' < random_id() < 'zzzzzzzz'
	True
	'''
	return ''.join(random.choices(string.ascii_letters, k=8))


if __name__ == '__main__':
	import doctest
	doctest.testmod()
